package com.sleepycat.je.txn;

import java.util.Map;

import com.sleepycat.je.dbi.MemoryBudget;
import com.sleepycat.je.txn.Txn.ReadLocksSet;
import com.sleepycat.je.txn.Txn.WriteInfoMap;

public privileged aspect TxnBudget {
	private final int READ_LOCK_OVERHEAD = MemoryBudget.HASHSET_ENTRY_OVERHEAD;

	private final int WRITE_LOCK_OVERHEAD = MemoryBudget.HASHMAP_ENTRY_OVERHEAD
			+ MemoryBudget.LONG_OVERHEAD;// MB

	after(TxnManager lm):
		call(Object Map.put(Object, Object)) && this(lm) {

		lm.env.getMemoryBudget().updateMiscMemoryUsage(
				MemoryBudget.HASHMAP_ENTRY_OVERHEAD);
	}

	after(TxnManager lm):
		call(Object *.remove(Object)) && this(lm) {

		lm.env.getMemoryBudget().updateMiscMemoryUsage(
				0 - MemoryBudget.HASHMAP_ENTRY_OVERHEAD);

	}

	private int Txn.inMemorySize;// MB

	/*
	 * accumluted memory budget delta. Once this exceeds ACCUMULATED_LIMIT we
	 * inform the MemoryBudget that a change has occurred.
	 */
	private int Txn.accumulatedDelta = 0;

	/*
	 * Max allowable accumulation of memory budget changes before MemoryBudget
	 * should be updated. This allows for consolidating multiple calls to
	 * updateXXXMemoryBudget() into one call. Not declared final so that unit
	 * tests can modify this. See SR 12273.
	 */
	public static int Txn.ACCUMULATED_LIMIT = 10000;

	before(Txn txn):
		execution(void TxnManager.registerTxn(Txn)) && args(txn) {

		/*
		 * Note: readLocks, writeInfo, undoDatabases, deleteDatabases are
		 * initialized lazily in order to conserve memory. WriteInfo and
		 * undoDatabases are treated as a package deal, because they are both
		 * only needed if a transaction does writes.
		 * 
		 * When a lock is added to this transaction, we add the collection entry
		 * overhead to the memory cost, but don't add the lock itself. That's
		 * taken care of by the Lock class.
		 */
		txn.updateMemoryUsage(MemoryBudget.TXN_OVERHEAD);// MB

	}

	private void Txn.updateMemoryUsage(int delta) {
		inMemorySize += delta;
		accumulatedDelta += delta;
		if (accumulatedDelta > Txn.ACCUMULATED_LIMIT
				|| accumulatedDelta < -Txn.ACCUMULATED_LIMIT) {
			envImpl.getMemoryBudget().updateMiscMemoryUsage(accumulatedDelta);
			accumulatedDelta = 0;
		}
	}

	int Txn.getAccumulatedDelta() {
		return accumulatedDelta;
	}

	int Txn.getInMemorySize() {// MB
		return inMemorySize;
	}

	after(Txn txn):
		call(boolean *.add(Object)) && target(ReadLocksSet) && this(txn) {

		txn.updateMemoryUsage(READ_LOCK_OVERHEAD);// MB
	}

	after(Txn txn):
		call(boolean *.remove(Object)) && target(ReadLocksSet) && this(txn) {

		txn.updateMemoryUsage(0 - READ_LOCK_OVERHEAD);// MB

	}

	after(Txn txn):
		call(ReadLocksSet.new()) && this(txn) {

		txn.updateMemoryUsage(MemoryBudget.HASHMAP_OVERHEAD);// MB
	}

	// TODO dynamic tests, advise more potential calls than necessary. this is
	// not homogenic!
	// this is even hard to figure out manually where to match.
	after(Txn txn):
		call(Object *.put(Object, Object)) && target(WriteInfoMap) && this(txn) {

		txn.updateMemoryUsage(WRITE_LOCK_OVERHEAD);// MB
	}

	after(Txn txn):
		call(Object *.remove(Object)) && target(WriteInfoMap) && this(txn) {

		txn.updateMemoryUsage(0 - WRITE_LOCK_OVERHEAD);// MB
	}

	after(Txn txn):
		call(WriteInfoMap.new()) && this(txn) {

		txn.updateMemoryUsage(MemoryBudget.TWOHASHMAPS_OVERHEAD);// MB
	}
}
